@@ -17,12 +17,20 @@
  */
 
 #include "kerncompat.h"
-#include <sys/stat.h>
+#include <errno.h>
+#include <string.h>
+#include "kernel-lib/bitops.h"
+#include "kernel-shared/accessors.h"
+#include "kernel-shared/extent_io.h"
+#include "kernel-shared/uapi/btrfs.h"
+#include "kernel-shared/uapi/btrfs_tree.h"
 #include "kernel-shared/ctree.h"
-#include "kernel-shared/disk-io.h"
-#include "kernel-shared/transaction.h"
 #include "kernel-shared/compression.h"
-#include "common/utils.h"
+#include "kernel-shared/file-item.h"
+#include "common/internal.h"
+#include "common/messages.h"
+
+struct btrfs_trans_handle;
 
 /*
  * Get the first file extent that covers (part of) the given range
@@ -186,7 +194,7 @@ int btrfs_read_file(struct btrfs_root *root, u64 ino, u64 start, int len,
 {
 	struct btrfs_fs_info *fs_info = root->fs_info;
 	struct btrfs_key key;
-	struct btrfs_path path;
+	struct btrfs_path path = { 0 };
 	struct extent_buffer *leaf;
 	struct btrfs_inode_item *ii;
 	u64 isize;
@@ -202,10 +210,9 @@ int btrfs_read_file(struct btrfs_root *root, u64 ino, u64 start, int len,
 		return -EINVAL;
 	}
 
-	btrfs_init_path(&path);
 	key.objectid = ino;
-	key.offset = start;
 	key.type = BTRFS_EXTENT_DATA_KEY;
+	key.offset = start;
 
 	ret = btrfs_search_slot(NULL, root, &key, &path, 0, 0);
 	if (ret < 0)
@@ -310,8 +317,8 @@ next:
 	if (no_holes) {
 		btrfs_release_path(&path);
 		key.objectid = ino;
-		key.offset = 0;
 		key.type = BTRFS_INODE_ITEM_KEY;
+		key.offset = 0;
 		ret = btrfs_lookup_inode(NULL, root, &path, &key, 0);
 		if (ret < 0)
 			goto out;
